#define PYFT2_MODULE
#include "pyft2.h"

#define FACE_FUNC(s, c) PyCFunction_New(&Face_methods[c], (PyObject*)s)



static PyObject *
Face_setCharSize(Face *self, PyObject *args, PyObject *kwds) {
	static char *kwlist[]={ "width", "height", "hres", "vres", NULL };
	FT_F26Dot6  char_width  = 0;
	FT_F26Dot6  char_height = 0;
	FT_UInt     horz_resolution = 0;
	FT_UInt     vert_resolution = 0;
	
	if(!PyArg_ParseTupleAndKeywords(args, kwds,
					"|iiii", kwlist,
					&char_width, &char_height,
					&horz_resolution, &vert_resolution)) {
		return NULL;
	}
	
	if(FT_Set_Char_Size(self->face,
			    char_width, char_height,
			    horz_resolution, vert_resolution)) {
		PyErr_SetString(PyExc_SystemError, "FT_Set_Char_Size failed");
		return NULL;
	}
	
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
Face_setPixelSizes(Face *self, PyObject *args, PyObject *kwds) {
	static char *kwlist[]={ "width", "height", NULL };
	FT_UInt  pixel_width  = 0;
	FT_UInt  pixel_height = 0;
	
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|ii", kwlist,
					&pixel_width, &pixel_height)) {
		return NULL;
	}

	if(FT_Set_Pixel_Sizes(self->face,
			      pixel_width, pixel_height)) {
		PyErr_SetString(PyExc_SystemError, "FT_Set_Pixel_Sizes failed");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;


}

static PyObject *
Face_attachFile(Face *self, PyObject *args) {
	char * s;
	if(!PyArg_ParseTuple(args, "s", &s)) {
		return NULL;
	}

	if(FT_Attach_File(self->face,s)) {
		PyErr_SetString(PyExc_SystemError, "FT_Attach_File failed");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
Face_setTransform(Face *self, PyObject *args, PyObject *kwds) {
	static char *kwlist[]={ "xx", "xy", "yx", "yy","dx","dy", NULL };
	FT_Matrix m = { 1 << 16, 0, 0, 1 << 16};
	FT_Vector v = {0, 0};
	
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|iiiiii", kwlist,
					&m.xx, &m.xy, &m.yx, &m.yy,
					&v.x, &v.y))
		return NULL;
	
	FT_Set_Transform(self->face, &m, &v);

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
Face_loadGlyph(Face *self, PyObject *args) {
	FT_UInt idx;
	int flags;
	flags = FT_LOAD_DEFAULT;
	if(!PyArg_ParseTuple(args, "i|i", &idx,&flags)) {
		return NULL;
	}

	if(FT_Load_Glyph(self->face, idx, flags)) {
		PyErr_SetString(PyExc_SystemError, "FT_Load_Glyph failed");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
Face_loadChar(Face *self, PyObject *args)
{
	int idx;
	int flags;
	flags = FT_LOAD_DEFAULT;
	if(!PyArg_ParseTuple(args, "i|i", &idx,&flags)) {
		return NULL;
	}

	if(FT_Load_Char(self->face, (FT_ULong)idx, flags)) {
		PyErr_SetString(PyExc_SystemError, "FT_Load_Char failed");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
Face_getCharIndex(Face *self, PyObject *args) {
	int c;
	if(!PyArg_ParseTuple(args, "i", &c))
		return NULL;
	return PyInt_FromLong((long)FT_Get_Char_Index(self->face,
						      (FT_ULong)c));
}

static PyObject *
Face_getNameIndex(Face *self, PyObject *args) {
	char * s;
	if(!PyArg_ParseTuple(args, "s", &s))
		return NULL;
	return PyInt_FromLong((long)FT_Get_Name_Index(self->face, s));
}

static PyObject *
Face_getGlyphName(Face *self, PyObject *args) {
#ifdef FT_CONFIG_OPTION_NO_GLYPH_NAMES
	char buf[80];
	FT_UInt i;
	if(!PyArg_ParseTuple(args, "i", &i))
		return NULL;
	if(!FT_Get_Glyph_Name(self->face, i, buf, sizeof(buf))) {
		PyErr_SetString(PyExc_SystemError, "FT_Get_Glyph_Name failed");
		return NULL;
	}
	return PyString_FromFormat("%s",buf);
#else	
	PyErr_SetString(PyExc_SystemError,
			"FT_Get_Glyph_Name was not compiled in.");
	return NULL;
#endif
}

static PyObject *
Face_getKerning(Face *self, PyObject *args) {
	FT_UInt l,r,m;
	FT_Vector v;
	m=0;
	if(!PyArg_ParseTuple(args, "ii|i", &l,&r,&m))
		return NULL;
	if(!FT_Get_Kerning(self->face, l,r,m,&v)) {
		PyErr_SetString(PyExc_SystemError, "FT_Get_Kerning failed");
		return NULL;
	}
	return Py_BuildValue("(ii)", v.x, v.y);
}

#if 0	/* Defined but unused */
static PyObject *
Face_renderGlyph(Face *self, PyObject *args, PyObject *kwds) {
	int s = 0;
	FT_GlyphSlot  slot = 0;
	FT_UInt       mode = 0; /* XXX change this */
	static char *kwlist[]={ "mode", "slot", NULL };
	
	mode = 0;
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|ii", kwlist,
					&mode,&s))
		return NULL;
	if(s<0) {
		PyErr_SetString(PyExc_IndexError,
				"Slot index must be positive");
		return NULL;
	}
	slot = self->face->glyph;
	while(s--) {
		if(slot->next == NULL) {
			PyErr_SetString(PyExc_IndexError,
					"Slot index to high");
			return NULL;
		}
		slot=slot->next;
	}
	if(!FT_Render_Glyph(slot,mode)) {
		PyErr_SetString(PyExc_SystemError, "FT_Render_Glyph failed");
		return NULL;
	}
	Py_INCREF(Py_None);
	return Py_None;
}
#endif


static PyObject *
Face_Glyph(Face *self, PyObject *args, PyObject *kwds) {
	Glyph *glyph;
	int s = 0;
	FT_GlyphSlot  slot;
	
	if(!PyArg_ParseTuple(args, "|i", &s))
		return NULL;
	if(s<0) {
		PyErr_SetString(PyExc_IndexError,
				"Slot index must be positive");
		return NULL;
	}
	
	slot = self->face->glyph;
	while(s--) {
		if(slot->next == NULL) {
			PyErr_SetString(PyExc_IndexError,"Slot index to high");
			return NULL;
		}
		slot = slot->next;
	}
	
	glyph = (Glyph*)((GlyphType.tp_alloc)(&GlyphType, 0));
	if (glyph == NULL) return NULL;
	glyph->refface = self;
	printf("Glyph_new() -> %p\n",glyph);
	if(FT_Get_Glyph(slot, &glyph->glyph)) {
		PyErr_SetString(PyExc_SystemError, "FT_Get_Glyph() failed");
		PyObject_Del(glyph);
		return NULL;
	}
	
	Py_XINCREF(self);
	return (PyObject*)glyph;
}


static PyObject *
Face_nextChar(Face *self, PyObject *args) {
	FT_UInt agindex;
	FT_ULong code, index;
	
	if(!PyArg_ParseTuple(args, "i", &code))
		return NULL;
	
	index = FT_Get_Next_Char(self->face, code, &agindex);
	return Py_BuildValue("(ii)",index, agindex);
}


enum {
	FACE_SETCHARSIZE,
	FACE_SETPIXELSIZES,
	FACE_ATTACHFILE,
	FACE_SETTRANSFORM,
	FACE_LOADCHAR,
	FACE_LOADGLYPH,
	FACE_GETCHARINDEX,
	FACE_GETNAMEINDEX,
	FACE_GETGLYPHNAME,
	FACE_GETKERNING,
	FACE_GLYPH,
	FACE_NEXTCHAR
};

static struct PyMethodDef Face_methods[] = {
	{ "setCharSize",
	  (PyCFunction)Face_setCharSize,   METH_VARARGS|METH_KEYWORDS,
	  (char*)NULL  },
	{ "setPixelSizes",
	  (PyCFunction)Face_setPixelSizes, METH_VARARGS|METH_KEYWORDS,
	  (char*)NULL  },
	{ "attachFile",
	  (PyCFunction)Face_attachFile, METH_VARARGS, (char*)NULL  },
	{ "setTransform",
	  (PyCFunction)Face_setTransform, METH_VARARGS|METH_KEYWORDS,
	  (char*)NULL  },
	{ "loadChar",
	  (PyCFunction)Face_loadChar, METH_KEYWORDS, (char*)NULL  },
	{ "loadGlyph",
	  (PyCFunction)Face_loadGlyph, METH_KEYWORDS, (char*)NULL  },
	{ "getCharIndex",
	  (PyCFunction)Face_getCharIndex, METH_VARARGS, (char*)NULL  },
	{ "getNameIndex",
	  (PyCFunction)Face_getNameIndex, METH_VARARGS, (char*)NULL  },
	{ "getGlyphName",
	  (PyCFunction)Face_getGlyphName, METH_VARARGS, (char*)NULL  },
	{ "getKerning",
	  (PyCFunction)Face_getKerning, METH_VARARGS, (char*)NULL  },
	{ "Glyph",
	  (PyCFunction)Face_Glyph, METH_VARARGS|METH_KEYWORDS, (char*)NULL  },
	{ "nextChar",
	  (PyCFunction)Face_nextChar, METH_VARARGS, (char*)NULL  },
	{ NULL, NULL, NULL, NULL }
};



static PyObject*
Face_getattr(Face*self, char*name) {
	PyObject *r;
	int i;
	FT_ULong index;
	FT_UInt  agindex;
	
	switch (name[0]) {
	case 'G':
		if(!strcmp(name,"Glyph"))
			return FACE_FUNC(self, FACE_GLYPH);
		break;
	case 'a':
		if(!strcmp(name,"availableSizes")) {
			r = PyList_New(self->face->num_fixed_sizes);
			if(r) for(i=0;i<self->face->num_fixed_sizes;i++)
				PyList_SET_ITEM(r,i,Py_BuildValue("(iiiii)",
				      self->face->available_sizes[i].height,
				      self->face->available_sizes[i].width,
				      self->face->available_sizes[i].size,
				      self->face->available_sizes[i].x_ppem,
				      self->face->available_sizes[i].y_ppem));
			return r;
		}
		if(!strcmp(name,"attachFile")) {
			return FACE_FUNC(self, FACE_ATTACHFILE);
		}
		if(self->face->face_flags & FT_FACE_FLAG_SCALABLE) {
			if(!strcmp(name,"ascender"))
			  return PyInt_FromLong((long)self->face->ascender);
		}
		break;
	case 'b':
		if(self->face->face_flags & FT_FACE_FLAG_SCALABLE) {
			if(!strcmp(name,"bbox"))
				return 	Py_BuildValue("(iiii)",
						      self->face->bbox.xMin,
						      self->face->bbox.yMin,
						      self->face->bbox.xMax,
						      self->face->bbox.yMax);
		}
		/* If the glyph is loaded and rendered as a bimap */
		/* */
		if(self->face->glyph->format == FT_GLYPH_FORMAT_BITMAP)  {
			if(!strcmp(name,"bitmap")) {
				return Py_BuildValue(
					"(iiii)",
					self->face->glyph->bitmap_left,
					self->face->glyph->bitmap_top,
					self->face->glyph->bitmap.width,
					self->face->glyph->bitmap.rows);
			}
		}   
		break;
	case 'c':
		if(!strcmp(name,"charmaps")) {
			r = PyList_New(self->face->num_charmaps);
			if(r)
			  for(i=0;i<self->face->num_charmaps;i++)
				  PyList_SET_ITEM(r,i,Py_BuildValue("(iii)",
				      self->face->charmaps[i]->encoding,
				      self->face->charmaps[i]->platform_id,
				      self->face->charmaps[i]->encoding_id));
			return r;
		}
		if(!strcmp(name,"charmap"))
			return Py_BuildValue(
				"(iii)",
				self->face->charmap->encoding,
				self->face->charmap->platform_id,
				self->face->charmap->encoding_id);
		break;
	case 'd':
		if(self->face->face_flags & FT_FACE_FLAG_SCALABLE) {
			if(!strcmp(name,"descender"))
			  return PyInt_FromLong((long)self->face->descender);
		}
		break;
	case 'f':
		if(!strcmp(name,"faceIndex"))
			return PyInt_FromLong((long)self->face->face_index);
		if(!strcmp(name,"faceFlags"))
			return PyInt_FromLong((long)self->face->face_flags);
		if(!strcmp(name,"familyName"))
			return PyString_FromString(self->face->family_name);
		if(!strcmp(name,"firstChar")) {
			index = FT_Get_First_Char(self->face, &agindex);
			return Py_BuildValue("(ii)",index, agindex);
		}
		break;
	case 'g':
		if(!strcmp(name,"getCharIndex"))
			return FACE_FUNC(self, FACE_GETCHARINDEX);
		if(!strcmp(name,"getNameIndex"))
			return FACE_FUNC(self, FACE_GETNAMEINDEX);
		if(!strcmp(name,"getKerning"))
			return FACE_FUNC(self, FACE_GETKERNING);
		if(!strcmp(name,"getGlyphName"))
			return FACE_FUNC(self, FACE_GETGLYPHNAME);
		if(!strcmp(name,"glyphFormat"))
			return PyInt_FromLong((long)self->face->glyph->format);
		break;
	case 'h':
		if(self->face->face_flags & FT_FACE_FLAG_SCALABLE)
			if(!strcmp(name,"height"))
			    return PyInt_FromLong((long)self->face->height);
		break;
	case 'l':
		if(!strcmp(name,"loadGlyph"))
			return FACE_FUNC(self, FACE_LOADGLYPH);
		if(!strcmp(name,"loadChar"))
			return FACE_FUNC(self, FACE_LOADCHAR);
		break;
	case 'm':
		if(self->face->face_flags & FT_FACE_FLAG_SCALABLE) {
		 if(!strcmp(name,"maxAdvanceWidth"))
		  return PyInt_FromLong((long)self->face->max_advance_width);
		 if(!strcmp(name,"maxAdvanceHeight"))
		  return PyInt_FromLong((long)self->face->max_advance_height);
		}
		break;
	case 'n':
		if(!strcmp(name,"nextChar"))
			return FACE_FUNC(self, FACE_NEXTCHAR);
		if(!strcmp(name,"numFaces"))
		     return PyInt_FromLong((long)self->face->num_faces);
		if(!strcmp(name,"numGlyphs"))
		     return PyInt_FromLong((long)self->face->num_glyphs);
		if(!strcmp(name,"numFixedSizes"))
		     return PyInt_FromLong((long)self->face->num_fixed_sizes);
		if(!strcmp(name,"numCharmaps"))
		     return PyInt_FromLong((long)self->face->num_charmaps);
		break;
	case 'p':
		if(!strcmp(name,"postscriptName"))
			/* Make sure that the string will not be deallocated */
			return PyString_FromFormat(
				"%s", FT_Get_Postscript_Name(self->face));
		break;
	case 's':
		if(!strcmp(name,"setCharSize"))
			return FACE_FUNC(self, FACE_SETCHARSIZE);
		if(!strcmp(name,"setPixelSizes"))
			return FACE_FUNC(self, FACE_SETPIXELSIZES);
		if(!strcmp(name,"setTransform"))
			return FACE_FUNC(self, FACE_SETTRANSFORM);
		if(!strcmp(name,"styleFlags"))
			return PyInt_FromLong((long)self->face->style_flags);
		if(!strcmp(name,"styleName"))
			return PyString_FromString(self->face->style_name);
		if(self->face->face_flags & FT_FACE_FLAG_SCALABLE)
		   if(!strcmp(name,"sizeMetrics"))
		      return Py_BuildValue(
			      "(iiiiiiii)",
			      self->face->size->metrics.x_ppem,
			      self->face->size->metrics.y_ppem,
			      self->face->size->metrics.x_scale,
			      self->face->size->metrics.y_scale,
			      self->face->size->metrics.ascender,
			      self->face->size->metrics.descender,
			      self->face->size->metrics.height,
			      self->face->size->metrics.max_advance);
		break;
	case 'u':
		if(self->face->face_flags & FT_FACE_FLAG_SCALABLE) {
		 if(!strcmp(name,"unitsPerEM"))
		  return PyInt_FromLong((long)self->face->units_per_EM);
		 if(!strcmp(name,"underlinePosition"))
		  return PyInt_FromLong((long)self->face->underline_position);
		 if(!strcmp(name,"underlineThickness"))
		  return PyInt_FromLong((long)self->face->underline_thickness);
		}
		break;
	default:
		break;
	}
	
	PyErr_SetString(PyExc_AttributeError, name);
	return NULL;
}

static int
Face_setattr(Face*self, char*name, PyObject * value) {
	int c;

	switch(name[0]) {		
	case 'c':
		if(!strcmp(name,"charmap")) {
			if (!PyInt_Check(value)) {
				PyErr_SetString(PyExc_TypeError,
						"Need an index");
				return -1;
			}
			c=PyInt_AsLong(value);		
			if (c>=self->face->num_charmaps || c<0) {
				PyErr_SetString(PyExc_IndexError,
						"Index error");
				return -1;
			}
			if(FT_Set_Charmap(self->face,
					  self->face->charmaps[c])) {
				PyErr_SetString(PyExc_SystemError,
						"FT_Set_Charmap failed");
				return -1;
			}
			return 0;
		}
		break;
	case 'e':
		if(!strcmp(name,"encoding")) {
			if (!PyInt_Check(value)) {
				PyErr_SetString(PyExc_TypeError,
						"Need an interger");
				return -1;
			}
			c=PyInt_AsLong(value);
			if(FT_Select_Charmap(self->face, c)) {
				PyErr_SetString(PyExc_SystemError,
						"FT_Select_Charmap failed");
				return -1;
			}
			return 0;
		}
		break;
	}
	PyErr_SetString(PyExc_AttributeError, name);
	return -1;
}

static void
Face_dealloc(Face *self) {
	printf("Face_dealloc() -> %p\n",self);
	FT_Done_Face(self->face);
	Py_DECREF((PyObject*)(self->reflib));
	PyObject_Del(self);
}

static PyObject*
Face_str(Face * self) {
	return PyString_FromFormat("<Face at %p>", self);
}

static PyObject*
Face_repr(Face * self) {
	return PyString_FromString("Face()");
}

PyTypeObject FaceType = {
	PyObject_HEAD_INIT(NULL)
	0,                          /* ob_size         */
	"GGI._ft2module.Face",     /* tp_name         */
	sizeof(Face),              /* tp_basicsize    */
	0,                          /* tp_itemsize     */
	(destructor)Face_dealloc,  /* tp_dealloc      */
	0,                          /* tp_print        */
	(getattrfunc)Face_getattr, /* tp_getattr      */
	(setattrfunc)Face_setattr, /* tp_setattr      */
	0,                          /* tp_compare      */
	(reprfunc)Face_repr,       /* tp_repr         */
	0,                          /* tp_as_number    */
	0,                          /* tp_as_sequence  */
	0,                          /* tp_as_mapping   */
	0,                          /* tp_hash         */
	0,                          /* tp_call         */
	(reprfunc)Face_str,        /* tp_str          */
	0,                          /* tp_getattro     */
	0,                          /* tp_setattro     */
	0,                          /* tp_as_buffer    */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,     /*tp_flags*/
	"Face object",            /* tp_doc          */
	
	/* 2.0 */
	0,		            /* tp_traverse     */
	0,		            /* tp_clear        */
	/* 2.1 */
	0,		            /* tp_richcompare  */
	0,		            /* tp_weaklistoffset */
	/* 2.2 */
	0,		            /* tp_iter         */
	0,		            /* tp_iternext     */
	0,                          /* tp_methods      */
	0,                          /* tp_members      */
	0,                          /* tp_getset       */
	0,                          /* tp_base         */
	0,                          /* tp_dict         */
	0,                          /* tp_descr_get    */
	0,                          /* tp_descr_set    */
	0,                          /* tp_dictoffset   */
	(initproc)0,       /* tp_init          */
	0,                          /* tp_alloc        */
	0, /* Only instanciated by library */ /* tp_new     */
	0,                          /* tp_is_gc        */
	0,                          /* tp_bases        */
        0,                          /* tp_mro          */
	0,                          /* tp_cache        */
	0,                          /* tp_subclasses   */
	0,                          /* tp_weaklist     */
	0                           /* tp_del          */
};
