#define PYFT2_MODULE
#include "pyft2.h"

#define LIBRARY_FUNC(s, c) PyCFunction_New(&Library_methods[c], (PyObject*)s)

static PyObject *
Library_Face(Library *self, PyObject *args, PyObject *kwds) {
	static char *kwlist[]={ "file", "index", NULL };
	char * file = NULL;
	int _index = 0;
	Face * face;
	
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "s|i", kwlist,
					&file, &_index))
		return NULL;
	
	face = (Face*)((FaceType.tp_alloc)(&FaceType, 0));
	if (face == NULL) return NULL;
	face->reflib = self;
	
	printf("Face_new() -> %p\n",face);
	
	if (FT_New_Face(self->library, file, _index, &face->face)) {
		PyErr_SetString(PyExc_SystemError,
				"FT_New_Face() failed");
		PyObject_Del(face);
		return NULL;
	}
	
	Py_XINCREF(self);
	return (PyObject*)face;
}

enum {
	LIBRARY_FACE = 0
};

static struct PyMethodDef Library_methods[] = {
	{ "Face",    (PyCFunction)Library_Face, METH_VARARGS|METH_KEYWORDS,
	  (char*)NULL  },
	{ NULL, NULL, 0, (char*)NULL }
};

static int
Library_init(Library * self, PyObject *args, PyObject *kwds) {
	printf("Library_new() -> %p\n",self);
	if (FT_Init_FreeType(&self->library)) {
		PyErr_SetString(PyExc_SystemError,"FT_Init_FreeType() failed");
		return -1;
	}
	return 0;
}

static void Library_dealloc(Library *self) {
	printf("Library_dealloc() -> %p\n",self);
	FT_Done_FreeType(self->library);
	self->ob_type->tp_free((PyObject*)self);
}

static PyObject *
Library_getattr(Library *self, char *name) {
	int a,b,c;
	switch(name[0]) {
	case 'v':
		if(!strcmp(name,"version")) {
			FT_Library_Version(self->library, &a, &b, &c);
			return Py_BuildValue("(iii)", a, b, c);
		}
		break;
	case 'F':
		if(!strcmp(name,"Face"))
			return LIBRARY_FUNC(self, LIBRARY_FACE);
		break;
	}
	PyErr_SetString(PyExc_AttributeError, name);
	return NULL;
}

static PyObject*
Library_str(Library * self) {
	return PyString_FromFormat("<Library at %p>",self->library);
}

static PyObject*
Library_repr(Library * self) {
	return PyString_FromString("Library()");
}

PyTypeObject LibraryType = {
	PyObject_HEAD_INIT(NULL)
	0,                          /* ob_size         */
	"GGI._ft2module.Library",   /* tp_name         */
	sizeof(Library),            /* tp_basicsize    */
	0,                          /* tp_itemsize     */
	(destructor)Library_dealloc,  /* tp_dealloc      */
	0,                            /* tp_print        */
	(getattrfunc)Library_getattr, /* tp_getattr      */
	(setattrfunc)0,              /* tp_setattr      */
	0,                          /* tp_compare      */
	(reprfunc)Library_repr,       /* tp_repr         */
	0,                          /* tp_as_number    */
	0,                          /* tp_as_sequence  */
	0,                          /* tp_as_mapping   */
	0,                          /* tp_hash         */
	0,                          /* tp_call         */
	(reprfunc)Library_str,        /* tp_str          */
	0,                          /* tp_getattro     */
	0,                          /* tp_setattro     */
	0,                          /* tp_as_buffer    */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,     /*tp_flags*/
	"Library object",            /* tp_doc          */
	
	/* 2.0 */
	0,		            /* tp_traverse     */
	0,		            /* tp_clear        */
	/* 2.1 */
	0,		            /* tp_richcompare  */
	0,		            /* tp_weaklistoffset */
	/* 2.2 */
	0,		            /* tp_iter         */
	0,		            /* tp_iternext     */
	0,                          /* tp_methods      */
	0,                          /* tp_members      */
	0,                          /* tp_getset       */
	0,                          /* tp_base         */
	0,                          /* tp_dict         */
	0,                          /* tp_descr_get    */
	0,                          /* tp_descr_set    */
	0,                          /* tp_dictoffset   */
	(initproc)Library_init,       /* tp_init         */
	0,                          /* tp_alloc        */
	PyType_GenericNew,  /*Library_new, */        /* tp_new          */
	0,                          /* tp_is_gc        */
	0,                          /* tp_bases        */
        0,                          /* tp_mro          */
	0,                          /* tp_cache        */
	0,                          /* tp_subclasses   */
	0,                          /* tp_weaklist     */
	0                           /* tp_del          */
};
